--[[
This sample script shows how to use the cda module (and its submodules) to
generate a CDA document.  Data that is hardcoded in this project would instead
be generate from another data source (a database, for example).
--]]

require('cda')

local function trace(A,B,C) end

header = {}
body = {}
body.pastMedicalHistory = {}
body.medications = {}
body.allergies = {}
body.familyHistory = {}
body.socialHistory = {}
body.physicalExam = {}
body.labs = {}
body.inofficeProcedure = {}
body.problems = {}
body.plan = {}

local DOCUMENT_ID = '2.16.840.1.113883.19.4'
local SET_ID = '2.16.840.1.113883.19.7'
local ORGANIZATION_ID = '2.16.840.1.113883.19.5'

-- Override default XML header
cda.XML_HEADER = [[<?xml version="1.0"?>
<?xml-stylesheet type="text/xsl" href="CDA.xsl"?>
]]

local MyLocalCodeSystem = {id='2.16.840.1.113883.19.278.47', name='MyLocalCodeSystem'}

-- The main function is the first function called from Iguana.
-- The Data argument will contain the message to be processed.
function main()   
   local Success = true
   local CDA
   
   if iguana.isTest() then
      Cda = GenerateCDA('12345')
   else   
      Success, Cda = pcall(GenerateCDA, '12345')
   end
   
   -- Easy way to validate: go to
   -- http://hit-testing.nist.gov/cda-validation/validation.html
   -- and select the file written out in the line below, then
   -- "Validate". Now, any time you want to see new validation
   -- results after running your script, simply reload the results
   -- page in your browser.  The browser should ask you if you want to
   -- resend the request, and when you say "yes", the new contents of
   -- the file will be validated.
   writeToFile('cda_out.xml', Cda)
   
   if not Success then
      net.http.respond{body=Cda, entity_type='text/plain'}
   else
      net.http.respond{body=Cda, entity_type='text/xml'}
   end
end

function GenerateCDA(PatientId)
   local Doc = cda.new()
   cda.setAttr(Doc, 'xmlns', 'voc', 'urn:hl7-org:v3/voc')
   
   -- Header
   header.document(Doc)
   header.patientInfo(Doc, PatientId)
   header.otherHeaderInfo(Doc)
   
   -- Body
   local Body = cda.createBody(Doc)
   
   -- History of Present Illness
   body.presentIllness(Body)
   
   -- Past Medical History
   local PMH = body.pastMedicalHistory.new(Body)
   body.pastMedicalHistory.asthma(PMH)
   body.pastMedicalHistory.hypertension(PMH)
   body.pastMedicalHistory.osteoarthritis(PMH)
   
   -- Medications
   -- This section better shows how one might generated
   -- the XML inside the section's <text>.
   local M, MT = body.medications.new(Body)
   local ML = MT:append(xml.ELEMENT, 'list')
   body.medications.theodur(M, ML)
   body.medications.proventil(M, ML)
   body.medications.prednisone(M, ML)
   body.medications.hctz(M, ML)
   
   -- Allergies & Adverse Reactions
   local AAR = body.allergies.new(Body)
   body.allergies.penicillin(AAR)
   body.allergies.aspirin(AAR)
   body.allergies.codeine(AAR)
   body.allergies.nausea(AAR)
   
   -- Family History
   local FH = body.familyHistory.new(Body)
   body.familyHistory.mi(FH)
   body.familyHistory.cancerAndDiabetes(FH)
   
   -- Social History
   local SH = body.socialHistory.new(Body)
   body.socialHistory.populate(SH)
   
   -- Physical Exam
   local PE = body.physicalExam.new(Body)
   body.physicalExam.vitalSigns(PE)
   body.physicalExam.skin(PE)
   body.physicalExam.lungs(PE)
   body.physicalExam.cardiac(PE)
   
   -- Labs
   local L = body.labs.new(Body)
   body.labs.imaging(L)
   body.labs.peakFlow(L)
   
   -- In-office Procedure
   local IOP, IOPT = body.inofficeProcedure.new(Body)
   local IOPL = IOPT:append(xml.ELEMENT, 'list')
   body.inofficeProcedure.sutureRemoval(IOP, IOPL)
   
   -- Problems
   local P = body.problems.new(Body)
   body.problems.asthma(P)
   body.problems.hyperTension(P)
   body.problems.contactDermatitis(P)
   
   -- Plan
   local Pl, PlT = body.plan.new(Body)
   local PlL = PlT:append(xml.ELEMENT, 'list')
   body.plan.pulmonary(Pl, PlL)
   body.plan.chem7(Pl, PlL)
   body.plan.teaching(Pl, PlL)
   body.plan.prednisone(Pl, PlL)
   body.plan.topical(Pl, PlL)
   body.plan.followup(Pl, PlL)
   
   return cda.serialize(Doc)
end

function header.document(Doc)
   local CD = Doc.ClinicalDocument
   cda.document.addTemplateId(CD, '2.16.840.1.113883.3.27.1776')
   cda.document.setId(CD, DOCUMENT_ID, 'c266')
   cda.document.setCode(CD, '11488-4', 'Consultation note')
   cda.document.setTitle(CD, 'Good Health Clinic Consultation Note')
   cda.document.setEffectiveTime(CD, 'April 7, 2000')
   cda.document.setConfidentialityCode(CD, 'N', 'Normal')
   cda.document.setSetId(CD, '2.16.840.1.113883.19.7', 'BB35')
   cda.document.setVersionNumber(CD, 2)
end

function header.patientInfo(Doc, Extension)
   local PR = cda.header.addPatientRole(Doc, ORGANIZATION_ID, Extension)
   cda.header.patient.addName(PR,  {family='Levin', given='Henry', suffix='the 7th'})
   cda.header.patient.setGender(PR, 'M')
   cda.header.patient.setBirthtime(PR, '24-Sep-1932')
   cda.header.patient.setProviderOrganization(PR, ORGANIZATION_ID)
end

function header.otherHeaderInfo(Doc)
   local DrRobert = {family='Dolin', given='Robert', suffix='MD'}
   
   local A = cda.header.addAuthor(Doc)
   cda.header.author.setTime(A, 'April 7, 2000 14:00:00', 4)
   local AA = cda.header.author.addAssignedAuthor(A, ORGANIZATION_ID, 'KP00017')
   cda.header.assigned.addName(AA, DrRobert)
   cda.header.assigned.addRepresentedOrganization(AA, ORGANIZATION_ID)
   
   local C = cda.header.addCustodian(Doc)
   cda.header.custodian.addAssignedCustodian(C, 'Good Health Clinic', ORGANIZATION_ID)
   
   local LA = cda.header.addAuthenticator(Doc)
   cda.header.authenticator.setTime(LA, 'April 8, 2000')
   cda.header.authenticator.setSignatureCode(LA, 'S')
   local LAAE = cda.header.authenticator.addAssignedEntity(LA, ORGANIZATION_ID, 'KP00017')
   cda.header.assigned.addName(LAAE, DrRobert)
   cda.header.assigned.addRepresentedOrganization(LAAE, ORGANIZATION_ID)
   
   local RD = cda.header.addParentDocument(Doc, 'RPLC')
   cda.document.setId(RD, DOCUMENT_ID, 'a123')
   cda.document.setSetId(RD, SET_ID, 'BB35')
   cda.document.setVersionNumber(RD, 1)
   
   local EE = cda.header.addEncompassingEncounter(Doc, '2.16.840.1.113883.19.6', 'KPENC1332')
   cda.header.encounter.setEffectiveTime(EE, 'April 07, 2000')
   local P = cda.header.encounter.addParticipant(EE, 'CON')
   cda.participant.setTime(P, 'April 07, 2000')
   local PAE = cda.participant.addAssignedEntity(P, ORGANIZATION_ID, 'KP00017')
   cda.header.assigned.addName(PAE, DrRobert)
   cda.header.assigned.addRepresentedOrganization(PAE, ORGANIZATION_ID)
   cda.header.encounter.setHealthCareFacility(EE, 'DSDLOC', 'GIM', 'General internal medicine clinic')
end

local HISTORY_OF_PRESENT_ILLNESS_TEXT =
[[<content styleCode="Bold">Henry Levin, the 7<sup>th</sup>
</content> is a 67 year old male referred for further asthma management. Onset of asthma in his <content revised="delete">twenties</content>
<content revised="insert">teens</content>. He was hospitalized twice last year, and already twice this year. He has not been able to be weaned off steroids for the past several months.]]

function body.presentIllness(Body)
   local HPI = cda.section.new(Body, '10164-2', 'History of Present Illness')
   local T = cda.section.addText(HPI)
   T:setInner(HISTORY_OF_PRESENT_ILLNESS_TEXT)
end

local PAST_MEDICAL_HISTORY_TEXT = 
[[<list>
	<item>
		<content ID="a1">Asthma</content>
	</item>
	<item>
		<content ID="a2">Hypertension (see HTN.cda for details)</content>
	</item>
	<item>
		<content ID="a3">Osteoarthritis, 
			<content ID="a4">right knee</content>
		</content>
	</item>
</list>]]

function body.pastMedicalHistory.new(Body)
   local PMH = cda.section.new(Body, '10153-2', 'Past Medical History')
   local T = cda.section.addText(PMH)
   T:setInner(PAST_MEDICAL_HISTORY_TEXT)
   
   return PMH
end

function body.pastMedicalHistory.asthma(S)   
   local O = cda.section.addObservation(S, 'COND', 'EVN')
   
   local C = cda.observation.addCode(O, '195967001', cda.cs.SNOMED_CT, 'Asthma')
   cda.observation.code.addOriginalTextReference(C, '#a1')
   
   cda.observation.setStatusCode(O, 'completed')
   cda.observation.setEffectiveTime(O, '1950')
   
   local R = cda.observation.addReference(O, 'XCRPT')
   cda.reference.setExternalObservation(R, '2.16.840.1.113883.19.1.2765')
   
   return O
end

function body.pastMedicalHistory.hypertension(S)
   local O = cda.section.addObservation(S, 'COND', 'EVN')
   
   local C = cda.observation.addCode(O, '59621000', cda.cs.SNOMED_CT, 'HTN')
   cda.observation.code.addOriginalTextReference(C, '#a2')
   
   cda.observation.setStatusCode(O, 'completed')
   
   local R1 = cda.observation.addReference(O, 'SPRT')
   cda.reference.setSeperatableInd(R1, false)
   local ED = cda.reference.addExternalDocument(R1)
   cda.document.setId(ED, DOCUMENT_ID..'.789')
   cda.document.addText(ED, 'multipart/related'):setInner('<reference value="HTN.cda"/>')
   cda.document.setSetId(ED, SET_ID..'.2465')
   cda.document.setVersionNumber(ED, 1)
   
   local R2 = cda.observation.addReference(O, 'XCRPT')
   cda.reference.setExternalObservation(R2, '2.16.840.1.113883.19.1.2005')
   
   return O
end

function body.pastMedicalHistory.osteoarthritis(S)
   local O = cda.section.addObservation(S, 'COND', 'EVN')
   
   local C = cda.observation.addCode(O, '396275006', cda.cs.SNOMED_CT, 'Osteoarthritis')
   cda.util.setXsiType(C, 'CD')
   cda.observation.code.addOriginalTextReference(C, '#a3')
   
   cda.observation.setStatusCode(O, 'completed')
   
   local TSC = cda.observation.addTargetSiteCode(O, '49076000', cda.cs.SNOMED_CT, 'Knee joint')
   cda.observation.code.addOriginalTextReference(TSC, '#a4')
   cda.observation.code.addQualifier(TSC,
      {code='78615007', displayName='with laterality', codeSystem=cda.cs.SNOMED_CT},
      {code='24028007', displayName='right', codeSystem=cda.cs.SNOMED_CT}
   )
   
   local R = cda.observation.addReference(O, 'XCRPT')
   cda.reference.setExternalObservation(R, '2.16.840.1.113883.19.1.1805')
   
   return O
end

local MEDICATIONS_TEXT =
[[<list>
	<item>Theodur 200mg BID</item>
	<item>Proventil inhaler 2puffs QID PRN</item>
	<item>Prednisone 20mg qd</item>
	<item>HCTZ 25mg qd</item>
</list>]]



function body.medications.new(Body)
   local S = cda.section.new(Body, '10160-0', 'Medications')
   return S, cda.section.addText(S)
end

local function addItem(L, Text)
   L:append(xml.ELEMENT, 'item'):append(xml.TEXT, Text)
end

function body.medications.theodur(S, L)
   addItem(L, 'Theodur 200mg BID')
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'EVN')
   cda.substanceadmin.addText(SA, 'Theodur 200mg BID')
   cda.substanceadmin.addEffectiveTime(SA, true, 12, 'h')
   cda.substanceadmin.setRouteCode(SA, 'PO', cda.cs.RouteOfAdministration)
   cda.substanceadmin.setDoseQuantity(SA, 200, 'mg')
   local MLD = cda.substanceadmin.addManufacturedLabeledDrug(SA)
   cda.substanceadmin.manufactured.addCode(
      MLD, '66493003', cda.cs.SNOMED_CT, 'Theophylline')
   return SA
end

function body.medications.proventil(S, L)
   addItem(L, 'Proventil inhaler 2puffs QID PRN')
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'EVN')
   cda.substanceadmin.addText(SA, 'Proventil inhaler 2puffs QID PRN')
   cda.substanceadmin.addEffectiveTime(SA, true, 6, 'h')
   cda.substanceadmin.setPriorityCode(SA, 'PRN')
   cda.substanceadmin.setRouteCode(SA, 'IPINHL', cda.cs.RouteOfAdministration, 'Inhalation, oral')
   cda.substanceadmin.setDoseQuantity(SA, 2)
   local MLD = cda.substanceadmin.addManufacturedLabeledDrug(SA)
   cda.substanceadmin.manufactured.addCode(
      MLD, '91143003', cda.cs.SNOMED_CT, 'Albuterol')
   return SA
end

function body.medications.prednisone(S, L)
   addItem(L, 'Prednisone 20mg qd')
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'EVN')
   cda.substanceadmin.setIdRoot(SA, '2.16.840.1.113883.19.8.1')
   cda.substanceadmin.addText(SA, 'Prednisone 20mg qd')
   cda.substanceadmin.addEffectiveTime(SA, true, 24, 'h')
   cda.substanceadmin.setRouteCode(SA, 'PO', cda.cs.RouteOfAdministration)
   cda.substanceadmin.setDoseQuantity(SA, 20, 'mg')
   local MLD = cda.substanceadmin.addManufacturedLabeledDrug(SA)
   cda.substanceadmin.manufactured.addCode(
      MLD, '10312003', cda.cs.SNOMED_CT, 'Prednisone preparation')
   return SA
end

function body.medications.hctz(S, L)
   addItem(L, 'HCTZ 25mg qd')
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'EVN')
   cda.substanceadmin.addText(SA, 'HCTZ 25mg qd')
   cda.substanceadmin.addEffectiveTime(SA, true, 24, 'h')
   cda.substanceadmin.setRouteCode(SA, 'PO', cda.cs.RouteOfAdministration)
   local MLD = cda.substanceadmin.addManufacturedLabeledDrug(SA)
   cda.substanceadmin.manufactured.addCode(
      MLD, '376209006', cda.cs.SNOMED_CT, 'Hydrochlorothiazide 25mg tablet')
   return SA
end

local ALLERGIES_TEXT =
[[<list>
	<item>Penicillin - Hives</item>
	<item>Aspirin - Wheezing</item>
	<item>Codeine - Itching and nausea</item>
</list>]]

function body.allergies.new(Body)
   local S = cda.section.new(Body, '10155-0', 'Allergies and Adverse Reactions')
   
   local T = cda.section.addText(S)
   T:setInner(ALLERGIES_TEXT)
   
   return S
end

function body.allergies.penicillin(S)
   local O = cda.section.addObservation(S, 'OBS', 'EVN')
   
   local C = cda.observation.addCode(O, '247472004', cda.cs.SNOMED_CT, 'Hives')
   cda.util.setXsiType(C, 'CD')
   
   cda.observation.setStatusCode(O, 'completed')
   
   local MO = cda.observation.addRelatedObservation(O, 'MFST', 'OBS', 'EVN')
   local MC = cda.observation.addCode(
      MO, '91936005', cda.cs.SNOMED_CT, 'Allergy to penicillin')
   cda.util.setXsiType(MC, 'CD')
   cda.observation.setStatusCode(MO, 'completed')
   
   return O
end

function body.allergies.aspirin(S)
   local O = cda.section.addObservation(S, 'OBS', 'EVN')
   
   cda.observation.addCode(O, '56018004', cda.cs.SNOMED_CT, 'Wheezing')
   cda.observation.setStatusCode(O, 'completed')
   
   local MO = cda.observation.addRelatedObservation(O, 'MFST', 'OBS', 'EVN')
   cda.observation.addCode(MO, '293586001', cda.cs.SNOMED_CT, 'Allergy to aspirin')
   cda.observation.setStatusCode(MO, 'completed')
   
   return O
end

function body.allergies.codeine(S)
   local O = cda.section.addObservation(S, 'OBS', 'EVN')
   
   cda.observation.addCode(O, '32738000', cda.cs.SNOMED_CT, 'Pruritis')
   cda.observation.setStatusCode(O, 'completed')
   
   local MO = cda.observation.addRelatedObservation(O, 'MFST', 'OBS', 'EVN')
   cda.observation.setIdRoot(MO, '2.16.840.1.113883.19.1.2010')
   local MC = cda.observation.addCode(
      MO, '62014003', cda.cs.SNOMED_CT, 'Adverse reaction to drug')
   
   cda.observation.code.addQualifier(MC,
      {code='246075003', codeSystem=cda.cs.SNOMED_CT, displayName='causative agent'},
      {code='1476002',   codeSystem=cda.cs.SNOMED_CT, displayName='codeine'})
   
   cda.observation.setStatusCode(MO, 'completed')
   
   return O
end

function body.allergies.nausea(S)
   local O = cda.section.addObservation(S, 'OBS', 'EVN')
   
   cda.observation.addCode(O, '73879007', cda.cs.SNOMED_CT, 'Nausea')
   cda.observation.setStatusCode(O, 'completed')
   
   local MO = cda.observation.addRelatedObservation(O, 'MFST', 'OBS', 'EVN')
   cda.observation.setIdRoot(MO, '2.16.840.1.113883.19.1.2010')
   cda.observation.addCode(MO, '84100007', cda.cs.SNOMED_CT)
   
   return O
end

local FAMILY_HISTORY_TEXT =
[[<list>
	<item>Father had fatal MI in his early 50's.</item>
	<item>No cancer or diabetes.</item>
</list>]]

function body.familyHistory.new(Body)
   local S = cda.section.new(Body, '10157-2', 'Family history')
   
   local T = cda.section.addText(S)
   T:setInner(FAMILY_HISTORY_TEXT)
   
   return S
end

function body.familyHistory.mi(S)
   local O = cda.section.addObservation(S, 'OBS', 'EVN')
   
   cda.observation.addCode(O, '22298006', cda.cs.SNOMED_CT, 'MI')
   cda.observation.setStatusCode(O, 'completed')
   cda.observation.setEffectiveTime(O, '1970')
   cda.observation.addRelatedSubject(O, 'PRS', 'FTH')
   
   local RO = cda.observation.addRelatedObservation(O, 'CAUS', 'OBS', 'EVN')
   cda.observation.addCode(RO, '399347008', cda.cs.SNOMED_CT, 'death')
   cda.observation.setStatusCode(RO, 'completed')
   cda.observation.setEffectiveTime(RO, '1970')
   
   return O
end

function body.familyHistory.cancerAndDiabetes(S)
   local CancO = cda.section.addObservation(S, 'OBS', 'EVN', true)
   cda.observation.addCode(
      CancO, '275937001', cda.cs.SNOMED_CT, 'Family history of cancer')
   cda.observation.setStatusCode(CancO, 'completed')
   local CancET = cda.observation.setEffectiveTime(CancO)
   cda.util.date.setRange(CancET, nil, {value='7 April 2000', inclusive=true})
   
   local DiabO = cda.section.addObservation(S, 'OBS', 'EVN')
   cda.observation.addCode(
      DiabO, '160274005', cda.cs.SNOMED_CT, 'No family history of diabetes')
   cda.observation.setStatusCode(DiabO, 'completed')
   local DiabET = cda.observation.setEffectiveTime(DiabO)
   cda.util.date.setRange(DiabET, nil, {value='7 April 2000', inclusive=true})
      
   return CancO, DiabO
end

local SOCIAL_HISTORY_TEXT =
[[<list>
	<item>Smoking :: 1 PPD between the ages of 20 and 55, and then he quit.</item>
	<item>Alcohol :: rare</item>
</list>]]

function body.socialHistory.new(Body)
   local S = cda.section.new(Body, '29762-2', 'Social History')
   
   local T = cda.section.addText(S)
   T:setInner(SOCIAL_HISTORY_TEXT)
   
   return S
end

function body.socialHistory.populate(S)
   local O1 = cda.section.addObservation(S, 'OBS', 'EVN')
   cda.observation.addCode(
      O1, '266924008', cda.cs.SNOMED_CT, 'ex-heavy cigarette smoker (20-39/day)')
   cda.observation.setStatusCode(O1, 'completed')
   local ET = cda.observation.setEffectiveTime(O1)
   cda.util.date.setRange(ET, {value='1955', precision=1}, {value='1990', precision=1})
   
   local O2 = cda.section.addObservation(S, 'OBS', 'EVN')
   cda.observation.addCode(
      O2, '160625004', cda.cs.SNOMED_CT, 'Date ceased smoking')
   cda.observation.setStatusCode(O2, 'completed')
   local V = cda.observation.addValue(O2, '1990', 'TS')
   
   local O3 = cda.section.addObservation(S, 'OBS', 'EVN')
   cda.observation.addCode(
      O3, '266917007', cda.cs.SNOMED_CT, 'Trivial drinker - less than 1/day')
   cda.observation.setStatusCode(O3, 'completed')
   
   return O1, O2
end

function body.physicalExam.new(Body)
   return cda.section.new(Body, '11384-5', 'Physical Examination')
end

local VITAL_SIGNS_TEXT =
[[<table>
	<tbody>
		<tr>
			<th>Date / Time</th>
			<th>April 7, 2000 14:30</th>
			<th>April 7, 2000 15:30</th>
		</tr>
		<tr>
			<th>Height</th>
			<td>177 cm (69.7 in)</td>
		</tr>
		<tr>
			<th>Weight</th>
			<td>194.0 lbs (88.0 kg)</td>
		</tr>
		<tr>
			<th>BMI</th>
			<td>28.1 kg/m2</td>
		</tr>
		<tr>
			<th>BSA</th>
			<td>2.05 m2</td>
		</tr>
		<tr>
			<th>Temperature</th>
			<td>36.9 C (98.5 F)</td>
			<td>36.9 C (98.5 F)</td>
		</tr>
		<tr>
			<th>Pulse</th>
			<td>86 / minute</td>
			<td>84 / minute</td>
		</tr>
		<tr>
			<th>Rhythm</th>
			<td>Regular</td>
			<td>Regular</td>
		</tr>
		<tr>
			<th>Respirations</th>
			<td>16 / minute, unlabored</td>
			<td>14 / minute</td>
		</tr>
		<tr>
			<th>Systolic</th>
			<td>132 mmHg</td>
			<td>135 mmHg</td>
		</tr>
		<tr>
			<th>Diastolic</th>
			<td>86 mmHg</td>
			<td>88 mmHg</td>
		</tr>
		<tr>
			<th>Position / Cuff</th>
			<td>Left Arm</td>
			<td>Left Arm</td>
		</tr>
	</tbody>
</table>]]

function body.physicalExam.vitalSigns(S)
   local SS = cda.section.new(S, '8716-3', 'Vital Signs')
   
   local T = cda.section.addText(SS)
   T:setInner(VITAL_SIGNS_TEXT)
   
   local HeightO = body.physicalExam.addVSObservation(
      SS, '50373000', 'Body height measure', '2000-04-07 14:30:00')
   local HeightV = cda.observation.addValue(HeightO, '1.77', 'PQ', 'm')
   cda.observation.value.addTranslationWithValue(
      HeightV, '69.7', '[in_I]', cda.cs.UCUM)
   
   local WeightO = body.physicalExam.addVSObservation(
      SS, '363808001', 'Body weight measure', '2000-04-07 14:30:00')
   local WeightV = cda.observation.addValue(WeightO, '194.0', 'PQ', '[lb_ap]')
   cda.observation.value.addTranslationWithValue(WeightV, '88.0', 'kg', cda.cs.UCUM)
   
   local BMIO = body.physicalExam.addVSObservation(
      SS, '60621009', 'Body mass index', '2000-04-07 14:30:00')
   cda.observation.addRtoPqPqValue(BMIO,
      {value='28.1', unit='kg'}, {value='1', unit='ar'})
   
   local BSAO = body.physicalExam.addVSObservation(
      SS, '301898006', 'Body surface area', '2000-04-07 14:30:00')
   cda.observation.addValue(BSAO, '2.05', 'PQ', 'ar')
   
   local BTO = body.physicalExam.addVSObservation(
      SS, '386725007', 'Body temperature', '2000-04-07 14:30:00')
   local BTV = cda.observation.addValue(BTO, '36.9', 'PQ', 'Cel')
   cda.observation.value.addTranslationWithValue(BTV, '98.5', '[degF]', cda.cs.UCUM)
   
   local HRO1 = body.physicalExam.addVSObservation(
      SS, '364075005', 'Heart rate', '2000-04-07 14:30:00')
   cda.observation.addRtoPqPqValue(HRO1, {value='86'}, {value='1', unit='min'})
   
   local HRO2 = body.physicalExam.addVSObservation(
      SS, '364075005', 'Heart rate', '2000-04-07 15:30:00')
   cda.observation.addRtoPqPqValue(HRO2, {value='84'}, {value='1', unit='min'})
   
   local RHRO1 = body.physicalExam.addVSObservation(
      SS, '364074009', 'Regularity of heart rhythm', '2000-04-07 14:30:00')
   cda.observation.addValueWithCode(RHRO1, '248649006', cda.cs.SNOMED_CT, 'Heart regular', 'CD')
   
   local RHRO2 = body.physicalExam.addVSObservation(
      SS, '364074009', 'Regularity of heart rhythm', '2000-04-07 15:30:00')
   cda.observation.addValueWithCode(RHRO2, '248649006', cda.cs.SNOMED_CT, 'Heart regular', 'CD')
   
   local RRO1 = body.physicalExam.addVSObservation(
      SS, '86290005', 'Respiratory rate', '2000-04-07 14:30:00')
   cda.observation.addRtoPqPqValue(RRO1, {value='16'}, {value='1', unit='min'})
   
   body.physicalExam.addVSObservation(
      SS, '276362002', 'Breathing easily', '2000-04-07 14:30:00')
   
   local RRO2 = body.physicalExam.addVSObservation(
      SS, '86290005', 'Respiratory rate', '2000-04-07 15:30:00')
   cda.observation.addRtoPqPqValue(RRO2, {value='14'}, {value='1', unit='min'})
   
   local BPO1 = body.physicalExam.addVSObservation(
      SS, '251076008', 'Cuff blood pressure', '2000-04-07 14:30:00')
   cda.observation.addTargetSiteCode(BPO1, '368208006', cda.cs.SNOMED_CT, 'Left arm')
   do
      local SBPO = cda.observation.addRelatedObservation(BPO1, 'COMP', 'OBS', 'EVN')
      cda.observation.addCode(SBPO, '271649006', cda.cs.SNOMED_CT, 'Systolic BP')
      cda.observation.setStatusCode(SBPO, 'completed')
      cda.observation.setEffectiveTime(SBPO, '2000-04-07 15:30:00', 5)
      cda.observation.addValue(SBPO, '132', 'PQ', 'mm[Hg]')
      
      local DBPO = cda.observation.addRelatedObservation(BPO1, 'COMP', 'OBS', 'EVN')
      cda.observation.addCode(DBPO, '271650006', cda.cs.SNOMED_CT, 'Diastolic BP')
      cda.observation.setStatusCode(DBPO, 'completed')
      cda.observation.setEffectiveTime(DBPO, '2000-04-07 15:30:00', 5)
      cda.observation.addValue(DBPO, '86', 'PQ', 'mm[Hg]')
   end
   
   local BPO2 = body.physicalExam.addVSObservation(
      SS, '251076008', 'Cuff blood pressure', '2000-04-07 15:30:00')
   cda.observation.addTargetSiteCode(BPO2, '368208006', cda.cs.SNOMED_CT, 'Left arm')
   do
      local SBPO = cda.observation.addRelatedObservation(BPO2, 'COMP', 'OBS', 'EVN')
      cda.observation.addCode(SBPO, '271649006', cda.cs.SNOMED_CT, 'Systolic BP')
      cda.observation.setStatusCode(SBPO, 'completed')
      cda.observation.setEffectiveTime(SBPO, '2000-04-07 15:30:00', 5)
      cda.observation.addValue(SBPO, '135', 'PQ', 'mm[Hg]')
      
      local DBPO = cda.observation.addRelatedObservation(BPO2, 'COMP', 'OBS', 'EVN')
      cda.observation.addCode(DBPO, '271650006', cda.cs.SNOMED_CT, 'Diastolic BP')
      cda.observation.setStatusCode(DBPO, 'completed')
      cda.observation.setEffectiveTime(DBPO, '2000-04-07 15:30:00', 5)
      cda.observation.addValue(DBPO, '88', 'PQ', 'mm[Hg]')
   end
   
   return SS
end

-- A wrapper around cda.section.addObservation which adds
-- common data for vital signs observations.
-- A good example of how you can wrap a few cda function calls
-- to create commonly used data.
function body.physicalExam.addVSObservation(S, Code, DisplayName, EffectiveTime)
   local O = cda.section.addObservation(S, 'OBS', 'EVN')	 	 
   cda.observation.addCode(O, Code, cda.cs.SNOMED_CT, DisplayName)	 	 
   cda.observation.setStatusCode(O, 'completed')	 	 
   cda.observation.setEffectiveTime(O, EffectiveTime, 5)	 	 
   return O	 	 
end

local SKIN_EXAM_TEXT =
[[Erythematous rash, palmar surface, left index finger.
<renderMultiMedia referencedObject="MM1"/>]]

function body.physicalExam.skin(S)
   local SS = cda.section.new(S, '8709-8', 'Skin Exam')
   
   local T = cda.section.addText(SS)
   T:setInner(SKIN_EXAM_TEXT)
   
   local O = cda.section.addObservation(SS, 'OBS', 'EVN')
   cda.observation.addCode(O, '271807003', cda.cs.SNOMED_CT, 'Rash')
   cda.observation.setStatusCode(O, 'completed')
   cda.observation.addMethodCode(
      O, '32750006', cda.cs.SNOMED_CT, 'Inspection')
   local TSC = cda.observation.addTargetSiteCode(
      O, '48856004', cda.cs.SNOMED_CT, 'Skin of palmer surface of index finger')
   cda.observation.code.addQualifier(TSC,
      {code='78615007', codeSystem=cda.cs.SNOMED_CT, displayName='with laterality'},
      {code='7771000',  codeSystem=cda.cs.SNOMED_CT, displayName='left'})
   
   local RI = cda.observation.addRegionOfInterest(
      O, 'SPRT', 'ROIOVL', 'EVN', 'MM1', '2.16.840.1.113883.19.3.1', 'ELLIPSE')
   
   cda.observation.addValue(RI, '3')
   cda.observation.addValue(RI, '1')
   cda.observation.addValue(RI, '3')
   cda.observation.addValue(RI, '7')
   cda.observation.addValue(RI, '2')
   cda.observation.addValue(RI, '4')
   cda.observation.addValue(RI, '4')
   cda.observation.addValue(RI, '4')
   
   cda.observation.addObservationMedia(
      RI, 'SUBJ', 'OBS', 'EVN', '2.16.840.1.113883.19.2.1', 'image/gif', 'lefthand.gif')
   
   return SS
end

function body.physicalExam.lungs(S)
   local SS = cda.section.new(S, '8710-6', 'Lungs')
   
   cda.section.addText(SS, 'Clear with no wheeze. Good air flow.')
   
   local O1 = cda.section.addObservation(SS, 'OBS', 'EVN')
   cda.observation.addCode(O1, '48348007', cda.cs.SNOMED_CT, 'Chest clear')
   cda.observation.setStatusCode(O1, 'completed')
   cda.observation.addMethodCode(O1, '37931006', cda.cs.SNOMED_CT, 'Auscultation')
   
   local O2 = cda.section.addObservation(SS, 'OBS', 'EVN', true)
   cda.observation.addCode(O2, '56018004', cda.cs.SNOMED_CT, 'Wheezing')
   cda.observation.setStatusCode(O2, 'completed')
   cda.observation.addMethodCode(O2, '37931006', cda.cs.SNOMED_CT, 'Auscultation')
   
   return SS
end

function body.physicalExam.cardiac(S)
   local SS = cda.section.new(S, '10223-2', 'Cardiac')
   
   cda.section.addText(SS, 'RRR with no murmur, no S3, no S4.')
   
   local HRO = cda.section.addObservation(SS, 'OBS', 'EVN')
   cda.observation.addCode(HRO, '76863003', cda.cs.SNOMED_CT, 'Normal heart rate')
   cda.observation.setStatusCode(HRO, 'completed')
   cda.observation.addMethodCode(HRO, '37931006', cda.cs.SNOMED_CT, 'Auscultation')
   
   local HMO = cda.section.addObservation(SS, 'OBS', 'EVN', true)
   cda.observation.addCode(HMO, '88610006', cda.cs.SNOMED_CT, 'heart murmur')
   cda.observation.setStatusCode(HMO, 'completed')
   cda.observation.addMethodCode(HMO, '37931006', cda.cs.SNOMED_CT, 'Auscultation')
   
   local S3O = cda.section.addObservation(SS, 'OBS', 'EVN', true)
   cda.observation.addCode(S3O, '277455002', cda.cs.SNOMED_CT, 'Third heart sound')
   cda.observation.setStatusCode(S3O, 'completed')
   cda.observation.addMethodCode(S3O, '37931006', cda.cs.SNOMED_CT, 'Auscultation')
   
   local S4O = cda.section.addObservation(SS, 'OBS', 'EVN')
   cda.observation.addCode(S4O, '60721002', cda.cs.SNOMED_CT, 'Fourth heart sound inaudible')
   cda.observation.setStatusCode(S4O, 'completed')
   cda.observation.addMethodCode(S4O, '37931006', cda.cs.SNOMED_CT, 'Auscultation')
   
   return SS
end

local LABS_TEXT =
[[<list>
	<item>CXR 02/03/1999: Hyperinflated. Normal cardiac silhouette, clear lungs.</item>					
	<item>Peak Flow today: 260 l/m</item>
</list>]]

function body.labs.new(Body)
   local S = cda.section.new(Body, '11502-2', 'Labs')
   
   local T = cda.section.addText(S)
   T:setInner(LABS_TEXT)
   
   return S
end

function body.labs.imaging(S)
   local O = cda.section.addObservation(S, 'OBS', 'EVN')
   
   cda.observation.addCode(O, '282290005', cda.cs.SNOMED_CT, 'Imaging interpretation')
   cda.observation.setStatusCode(O, 'completed')
   
   local COMP1 = cda.observation.addRelatedObservation(O, 'COMP', 'OBS', 'EVN')
   cda.observation.setIdRoot(COMP1, '2.16.840.1.113883.19.1.3005')
   cda.observation.addCode(COMP1, '249674001', cda.cs.SNOMED_CT, 'Chest hyperinflated')
   
   local COMP2 = cda.observation.addRelatedObservation(O, 'COMP', 'OBS', 'EVN')
   cda.observation.setIdRoot(COMP2, '2.16.840.1.113883.19.1.5505')
   local COMP2C = cda.observation.addCode(COMP2)
   cda.observation.code.addNullFlavor(COMP2C, 'OTH')
   cda.observation.code.addOriginalText(COMP2C, 'normal cardiac silhouette')
   
   local COMP3 = cda.observation.addRelatedObservation(
      O, 'COMP', 'OBS', 'EVN', {negationInd=true})
   cda.observation.setIdRoot(COMP3, '2.16.840.1.113883.19.1.6675')
   local COMP3C = cda.observation.addCode(COMP3)
   cda.observation.code.addNullFlavor(COMP3C, 'OTH')
   cda.observation.code.addOriginalText(COMP3C, 'radiopacities')
   
   local R = cda.observation.addReference(O, 'SPRT')
   local EO = cda.reference.setExternalObservation(
      R, '2.16.840.1.113883.19.1.14', 'DGIMG')
   cda.observation.addCode(EO, '56350004', cda.cs.SNOMED_CT, 'Chest-X-ray')
   
   return O
end

function body.labs.peakFlow(S)
   local O = cda.section.addObservation(S, 'OBS', 'EVN')
   
   cda.observation.setIdRoot(O, '2.16.840.1.113883.19.1.7005')
   cda.observation.addCode(O, '313193002', cda.cs.SNOMED_CT, 'Peak flow')
   cda.observation.setStatusCode(O, 'completed')
   cda.observation.setEffectiveTime(O, 'April 7 2000', 3)
   cda.observation.addRtoPqPqValue(O,
      {value=260, unit='l'}, {value=1, unit='min'})
   
   return O
end

function body.inofficeProcedure.new(Body)
   local S = cda.section.new(Body, '29554-3', 'In-office Procedures')
   return S, cda.section.addText(S)
end

function body.inofficeProcedure.sutureRemoval(S, L)
   addItem(L, 'Suture removal, left forearm.')
   
   local P = cda.section.addProcedureEntry(S, 'PROC', 'EVN')
   cda.entry.addCode(P, '30549001', cda.cs.SNOMED_CT, 'Suture removal')
   cda.entry.setStatusCode(P, 'completed')
   cda.entry.setEffectiveTime(P, '200004071430', 5)
   cda.entry.addTargetSiteCode(P, '66480008', cda.cs.SNOMED_CT, 'Left forearm')
   
   return P
end

local ASSESSMENT_TEXT =
[[<list>
	<item>Asthma, with prior smoking history. Difficulty weaning off steroids. Will try gradual taper.</item>
	<item>Hypertension, well-controlled.</item>
	<item>Contact dermatitis on finger.</item>
</list>]]

function body.problems.new(Body)
   local S = cda.section.new(Body, '11496-7', 'Assessment')
   
   local T = cda.section.addText(S)
   T:setInner(ASSESSMENT_TEXT)
   
   return S
end

function body.problems.asthma(S)
   local O = cda.section.addObservation(S, 'COND', 'EVN')
   
   cda.observation.addCode(O, '14657009', cda.cs.SNOMED_CT, 'Established diagnosis')
   cda.observation.setStatusCode(O, 'completed')
   cda.observation.setEffectiveTime(O, 'April 7 2000 3:30:00pm', 5)
   
   local V = cda.observation.addValueWithCode(O, '195967001', cda.cs.SNOMED_CT, 'Asthma', 'CD')
   cda.observation.value.addTranslationWithDisplayName(
      V, '49390', cda.cs.ICD9CM, 'ASTHMA W/O STATUS ASTHMATICUS')
   
   local R = cda.observation.addReference(O, 'ELNK')
   cda.reference.setExternalObservation(R, '2.16.840.1.113883.19.1.35', 'COND')
   
   return O
end

function body.problems.hyperTension(S)
   local O = cda.section.addObservation(S, 'COND', 'EVN')
   
   cda.observation.addCode(O, '14657009', cda.cs.SNOMED_CT, 'Established diagnosis')
   cda.observation.setStatusCode(O, 'completed')
   cda.observation.setEffectiveTime(O, '2000-04-07 15:30', 5)
   
   local V = cda.observation.addValueWithCode(O, '59621000', cda.cs.SNOMED_CT, 'Essential hypertension', 'CD')
   cda.observation.value.addTranslationWithDisplayName(
      V, '4019', cda.cs.ICD9CM, 'HYPERTENSION NOS')
   
   local R = cda.observation.addReference(O, 'ELNK')
   cda.reference.setExternalObservation(R, '2.16.840.1.113883.19.1.37', 'COND')
   
   return O
end

function body.problems.contactDermatitis(S)
   local O = cda.section.addObservation(S, 'COND', 'EVN')
   
   cda.observation.addCode(O, '14657009', cda.cs.SNOMED_CT, 'Established diagnosis')
   cda.observation.setStatusCode(O, 'completed')
   cda.observation.setEffectiveTime(O, '2000-04-07 15:30', 5)
   
   local V = cda.observation.addValueWithCode(O, '40275004', cda.cs.SNOMED_CT, 'Contact dermatitis', 'CD')
   cda.observation.value.addTranslationWithDisplayName(
      V, '692.9', cda.cs.ICD9CM, 'Contact Dermatitis, NOS')
   
   local TSC = cda.observation.addTargetSiteCode(O, '48856004',
      cda.cs.SNOMED_CT, 'Skin of palmer surface of index finger')
   cda.observation.code.addQualifier(TSC,
      {code='78615007', codeSystem=cda.cs.SNOMED_CT, displayName='with laterality'},
      {code='7771000',  codeSystem=cda.cs.SNOMED_CT, displayName='left'})
   
   return O
end

function body.plan.new(Body)
   local S = cda.section.new(Body, '18776-5', 'Plan', {root='2.16.840.1.113883.3.27.354'})
   return S, cda.section.addText(S)
end

function body.plan.pulmonary(S, L)
   addItem(L, 'Complete PFTs with lung volumes.')
   
   local A = cda.section.addActEntry(S, 'ACT', 'INT')
   A:append(xml.ELEMENT, 'id') -- empty id element, probably unnecessary
   cda.entry.addCode(A, '23426006', cda.cs.SNOMED_CT, 'Pulmonary function test')
   cda.entry.addText(A, 'Complete PFTs with lung volumes.')
   local AA = cda.entry.addRelatedAct(A, 'COMP', 'ACT', 'INT')
   cda.entry.addCode(AA, '252472004', cda.cs.SNOMED_CT, 'Lung volume test')
   
   return A
end

function body.plan.chem7(S, L)
   addItem(L, 'Chem-7 tomorrow.')
   
   local O = cda.section.addObservation(S, 'OBS', 'INT')
   
   local C = cda.observation.addCode(O, '24320-4')
   -- Different code system than the default
   C:setAttr('codeSystem', '2.16.840.1.113883.6.1')
   C:setAttr('codeSystemName', 'LOINC')
   cda.observation.code.addOriginalText(C, 'Chem-7')
   cda.observation.value.addTranslationWithDisplayName(
      C, 'aYU7t6', MyLocalCodeSystem, 'Chem7')
   cda.observation.addText(O, 'Chem-7 tomorrow')
   cda.observation.setEffectiveTime(O, 'April 8 2000', 3)
   
   return O
end

function body.plan.teaching(S, L)
   addItem(L, 'Teach peak flow rate measurement.')
   
   local A = cda.section.addActEntry(S, 'ACT', 'INT')
   A:append(xml.ELEMENT, 'id') -- empty id element, probably unnecessary
   local C = cda.entry.addCode(A, '223468009', cda.cs.SNOMED_CT, 'Teaching of skills')
   cda.entry.code.addQualifier(C,
      {code='363702006', codeSystem=cda.cs.SNOMED_CT, displayName='has focus'},
      {code='29893006',  codeSystem=cda.cs.SNOMED_CT, displayName='Peak flow rate measurement'})
   
   return A
end

function body.plan.prednisone(S, L)
   addItem(L, 'Decrease prednisone to 20qOD alternating with 18qOD.')
   
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'RQO')
   cda.substanceadmin.addText(SA, 'prednisone 20qOD alternating with 18qOD.')
   cda.substanceadmin.setRouteCode(SA, 'PO', cda.cs.RouteOfAdministration)
   local MLD = cda.substanceadmin.addManufacturedLabeledDrug(SA)
   cda.substanceadmin.manufactured.addCode(
      MLD, '10312003', cda.cs.SNOMED_CT, 'Prednisone preparation')
   
   return SA
end

function body.plan.topical(S, L)
   addItem(L, 'Hydrocortisone cream to finger BID.')
   
   local SA = cda.section.addSubstanceAdmin(S, 'SBADM', 'RQO')
   cda.substanceadmin.addText(SA, 'Hydrocortisone cream to finger BID.')
   cda.substanceadmin.addEffectiveTime(SA, true, '12', 'h')
   cda.substanceadmin.setRouteCode(SA, 'SKIN', cda.cs.RouteOfAdministration, 'Topical application, skin')
   local ASC = cda.substanceadmin.setApproachSiteCode(
      SA, '48856004', cda.cs.SNOMED_CT, 'Skin of palmer surface of index finger')
   cda.substanceadmin.code.addQualifier(ASC,
      {code='78615007', codeSystem=cda.cs.SNOMED_CT, displayName='with laterality'},
      {code='7771000',  codeSystem=cda.cs.SNOMED_CT, displayName='left'})
   local MLD = cda.substanceadmin.addManufacturedLabeledDrug(SA)
   cda.substanceadmin.manufactured.addCode(
      MLD, '331646005', cda.cs.SNOMED_CT, 'Hydrocortisone cream')
   
   return SA
end

function body.plan.followup(S, L)
   addItem(L, 'RTC 1 week.')
   
   local E = cda.section.addEncounterEntry(S, 'ENC', 'RQO')
   cda.entry.addCode(E, '185389009', cda.cs.SNOMED_CT, 'Follow-up visit')
   local ET = cda.entry.setEffectiveTime(E)
   cda.util.date.setRange(ET, {value='20000412'}, {value='20000417'})
end

function writeToFile(FilePath, FileContent)
   local CdaOut = io.open(FilePath, 'w')
   CdaOut:write(FileContent)
   CdaOut:close()
end
