cda.util = {}
cda.util.name = {}
cda.util.date = {}

imp = {}

function imp.checkParam(Args, Expected, Usage)
   if type(Args) ~= 'table' then
      error('Expected parameter to be a table\n'..Usage, 3)
   end
   for Key,_ in pairs(Args) do
      if not Expected[Key] then
         error('Unexpected parameter "'..Key..'"\n'..Usage, 3)
      end
   end
   for Key, Optionality in pairs(Expected) do
      debug(Key)
      if Optionality == 2 and Args[Key] == nil then
         error('Required parameter "'..Key..'" not present.\n'..Usage,3)
      end
   end
end

function imp.getOrCreateElement(Parent, Name)
   return Parent[Name] or Parent:append(xml.ELEMENT, Name)
end

-- Root and Extension are optional
function imp.getOrCreateChildWithId(E, ChildElementName, Root, Extension, AssigningAuthorityName)
   local C = imp.getOrCreateElement(E, ChildElementName)
   if Root or Extension then
      local I = imp.getOrCreateElement(C, 'id')
      imp.setIdAttributes(I, Root, Extension, AssigningAuthorityName)
   end
   return C
end

-- Root, Extension, and AssigningAuthorityName are optional
function imp.createChildWithId(E, ChildElementName, Root, Extension, AssigningAuthorityName)
   local C = E:append(xml.ELEMENT, ChildElementName)
   if Root or Extension or AssigningAuthorityName then
      local I = C:append(xml.ELEMENT, 'id')
      imp.setIdAttributes(I, Root, Extension, AssigningAuthorityName)
   end
   return C
end

function imp.setIdAttributes(I, Root, Extension, AssigningAuthorityName)
   if Root then
      I:setAttr('root', Root)
   end
   if Extension then
      I:setAttr('extension', Extension)
   end
   if AssigningAuthorityName then
      I:setAttr('assigningAuthorityName', AssigningAuthorityName)
   end
end

function cda.util.clear(Node)
   while #Node > 0 do
      Node:remove(#Node)
   end
   return Node
end

local SET_NAME_USAGE = [[
Add and populate a person name element in a CDA document 

Returns: The populated name element.
Accepts:
 1) An XML node to be populated with the person name information
 2) A table with the following optional entries
   'family'   - Family name - i.e. Smith
   'given'    - Given name - i.e. John
   'suffix'   - Suffix - like "the III"
   'prefix'   - Prefix - like "Mr."
 3) The value of the "use" attribute (optional).

For example:

cda.util.name.set(P.name, {family='Levin', given='Henry', suffix='the 7th'}, 'P')  
]]

function cda.util.addName(E, Arg, Use)
   imp.checkParam(Arg, {family=1, given=1, suffix=1, prefix=1}, SET_NAME_USAGE)
   local N = E:append(xml.ELEMENT, 'name')
   N:setAttr('use', Use)
   if Arg.prefix  then N:append(xml.ELEMENT, 'prefix'):append(xml.TEXT, Arg.prefix) end
   if Arg.given   then N:append(xml.ELEMENT, 'given'):append(xml.TEXT, Arg.given)   end
   if Arg.family  then N:append(xml.ELEMENT, 'family'):append(xml.TEXT, Arg.family) end
   if Arg.suffix  then N:append(xml.ELEMENT, 'suffix'):append(xml.TEXT, Arg.suffix) end
   return N
end

function cda.util.date.setDate(Node, FuzzyDate, Precision)
   local FormatString = '%Y%m%d%H%M%S'
   Precision = Precision or 3
   FormatString = FormatString:sub(1, Precision*2)
   
   Node:setAttr('value', os.date(FormatString, dateparse.parse(FuzzyDate))) 
   return Node, Node.value:nodeValue()
end

local SET_RANGE_USAGE = [[
Populate an element with a datetime range.

Returns: the populated element.
Accepts:
 1) An XML node to be populated with the datetime range.
 2) An optional table representing the "low end" of the range.
    This table can have the following optional entries:
    'value' - the fuzzy datetime value
    'inclusive' - boolean, is this end of the range inclusive?
    'precision' - the precision with which the datetime should be rendered.
    'nullFlavor' - the value of the nullFlavor attribute.
 3) An optional table representing the "high end" of the range.
    The table should be in the same format as the "low end" table.
]]

function cda.util.date.setRange(Node, Low, High)
   if Low then
      imp.checkParam(Low,
         {value=1, inclusive=1, precision=1, nullFlavor=1}, SET_RANGE_USAGE)
      local L = imp.getOrCreateElement(Node, 'low')
      if Low.value then
         cda.util.date.setDate(L, Low.value, Low.precision)
      end
      if Low.inclusive ~= nil then
         L:setAttr('inclusive', tostring(Low.inclusive))
      end
      if Low.nullFlavor then
         L:setAttr('nullFlavor', Low.nullFlavor)
      end
   end
   if High then
      imp.checkParam(High,
         {value=1, inclusive=1, precision=1, nullFlavor=1}, SET_RANGE_USAGE)
      local H = imp.getOrCreateElement(Node, 'high')
      if High.value then
         cda.util.date.setDate(H, High.value, High.precision)
      end
      if High.inclusive ~= nil then
         H:setAttr('inclusive', tostring(High.inclusive))
      end
      if High.nullFlavor then
         H:setAttr('nullFlavor', High.nullFlavor)
      end
   end
   return Node
end

-- A "coded element" refers to an element with the following attributes:
--  * code
--  * codeSystem (from CodeSystem.id)
--  * codeSystemName (from CodeSystem.name)
--  * displayName (optional)
--
function cda.util.addCodeElement(O, ElementName, Code, CodeSystem, DisplayName)
   local C = imp.getOrCreateElement(O, ElementName)
   if Code then
      C:setAttr('code', Code)
   end
   if CodeSystem then
      cda.codesystem.set(C, CodeSystem)
   end
   if DisplayName then
      C:setAttr('displayName', DisplayName)
   end
   return C
end

-- Works on any element
function cda.util.setXsiType(E, XsiType)
   if not E['xsi:type'] then
      E:insert(1, xml.ATTRIBUTE, 'xsi:type')
   end
   E:setAttr('xsi:type', XsiType)
   return E
end

-- TemplateIds should be a list of tables with the following
-- optional entries:
-- root, extension, assigningAuthorityName
-- For example:
-- cda.util.addTemplateIds(E, {
--    {root='1.3.6.1.4.1.19376.1.5.3.1.4.5.1'},
--    {root='1.3.6.1.4.1.19376.1.5.3.1.4.5.3'}
-- })
function cda.util.addTemplateIds(E, TemplateIds)
   for i,V in ipairs(TemplateIds) do
      local TI = E:append(xml.ELEMENT, 'templateId')
      if V.root then
         TI:setAttr('root', V.root)
      end
      if V.extension then
         TI:setAttr('extension', V.extension)
      end
      if V.assigningAuthorityName then
         TI:setAttr('assigningAuthorityName', V.assigningAuthorityName)
      end
   end
end
